/* Copyright (C) 2015-2021 |Meso|Star> (contact@meso-star.com)
 *
 * This software is a library whose purpose is to generate [pseudo] random
 * numbers and random variates.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "ssp_rng_c.h"

#include <rsys/dynamic_array_double.h>
#include <rsys/mem_allocator.h>
#include <rsys/ref_count.h>

using discrete_dist=RAN_NAMESPACE::discrete_distribution<size_t>;

struct ssp_ranst_discrete {
  struct darray_double pdf;
  ref_T ref;
  struct mem_allocator* allocator;
  discrete_dist* distrib;
};

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static void
discrete_release(ref_T* ref)
{
  struct ssp_ranst_discrete* ran;
  ASSERT(ref);
  ran = CONTAINER_OF(ref, struct ssp_ranst_discrete, ref);
  if(ran->distrib) {
    ran->distrib->~discrete_dist();
    MEM_RM(ran->allocator, ran->distrib);
  }
  darray_double_release(&ran->pdf);
  MEM_RM(ran->allocator, ran);
}

/*******************************************************************************
 * Exported functions
 ******************************************************************************/
res_T
ssp_ranst_discrete_create
  (struct mem_allocator* allocator,
   struct ssp_ranst_discrete** out_ran)
{
  struct ssp_ranst_discrete* ran = nullptr;
  void* mem = nullptr;
  res_T res = RES_OK;

  if(!out_ran) {
    res = RES_BAD_ARG;
    goto error;
  }
  allocator = allocator ? allocator : &mem_default_allocator;

  ran = static_cast<ssp_ranst_discrete*>
    (MEM_CALLOC(allocator, 1, sizeof(ssp_ranst_discrete)));
  if(!ran) {
    res = RES_MEM_ERR;
    goto error;
  }
  ref_init(&ran->ref);

  mem = MEM_ALLOC(allocator, sizeof(discrete_dist));
  if(!mem) {
    res = RES_MEM_ERR;
    goto error;
  }
  ran->allocator = allocator;
  ran->distrib = static_cast<discrete_dist*>(new(mem) discrete_dist);
  darray_double_init(allocator, &ran->pdf);

exit:
  if(out_ran) *out_ran = ran;
  return res;
error:
  if(ran) {
    SSP(ranst_discrete_ref_put(ran));
    ran = NULL;
  }
  goto exit;
}

res_T
ssp_ranst_discrete_setup
  (struct ssp_ranst_discrete* ran,
   const double* weights,
   const size_t nweights)
{
  double* pdf;
  double sum = 0;
  size_t i;
  res_T res = RES_OK;

  if(!ran || !weights || !nweights) {
    res = RES_BAD_ARG;
    goto error;
  }

  res = darray_double_resize(&ran->pdf, nweights);
  if(res != RES_OK) goto error;

  pdf = darray_double_data_get(&ran->pdf);

  FOR_EACH(i, 0, nweights) {
    if(weights[i] < 0.f) {
      res = RES_BAD_ARG;
      goto error;
    }
    sum += weights[i];
    pdf[i] = weights[i];
  }
  if(sum == 0) {
    res = RES_BAD_ARG;
    goto error;
  }

  sum = 1 / sum;
  FOR_EACH(i, 0, nweights) {
    pdf[i] *= sum;
  }

  {
    discrete_dist::param_type p{weights, weights + nweights};
    ran->distrib->param(p);
  }

exit:
  return res;
error:
  if(ran) {
    darray_double_clear(&ran->pdf);
  }
  goto exit;
}

res_T
ssp_ranst_discrete_ref_get(struct ssp_ranst_discrete* ran)
{
  if(!ran) return RES_BAD_ARG;
  ref_get(&ran->ref);
  return RES_OK;
}

res_T
ssp_ranst_discrete_ref_put(struct ssp_ranst_discrete* ran)
{
  if(!ran) return RES_BAD_ARG;
  ref_put(&ran->ref, discrete_release);
  return RES_OK;
}

size_t
ssp_ranst_discrete_get(struct ssp_rng* rng, const struct ssp_ranst_discrete* ran)
{
  class rng_cxx r(*rng);
  return ran->distrib->operator()(r);
}

double
ssp_ranst_discrete_pdf(const size_t i, const struct ssp_ranst_discrete* ran)
{
  ASSERT(ran && i < darray_double_size_get(&ran->pdf));
  return darray_double_cdata_get(&ran->pdf)[i];
}

